#include "defo.h"
#include "krawczyk-newton.h"
#include "my-interval-func.h"
#include "bodies.h"

#include "cc-nb.h"

using namespace std;
using namespace capd;

//#define TESTY_KRAWCZYK

// #define REFINE_CC_TESTY

/* ================= */
/*  Newton's method  */
/* ================= */

MyVector NewtonOperator(const MyMatrix &invDF, const Bodies &bodies)
//  returns: x0 - invDF * Fx0
{
  int dim = bodies.dimRedConfSpace();

  MyVector Fx0(dim);
  MyVector x0(dim);
  Bodies *midBodies = bodies.calculateMidBodies();
  x0 = midBodies->toReducedSpace();
  midBodies->compFRed(Fx0);
  delete midBodies;

  return x0 - invDF * Fx0;
}

ResultType newtonMethod(Bodies copyBodies, Bodies &returnValues, bool &intersec, MyMatrix &DFF)
{

  return ResultType::methodFailed;
}

/* =================== */
/*  Krawczyk's method  */
/* =================== */

MyVector krawczykOperator(const MyMatrix &C, const MyMatrix &DFF, const Bodies &bodies, MyVector &chZ)
// Input:
//   C - some linear isomorphism
//   DFF - D Freduced with respect to coordinates parameterizing the reduced space
//   bodies - the set on which the Krawczyk operator is computed
// on output:
//   returns  x0 - C * F(x0) + (I - C*DFF(x)) * (x-x0),
//       where x = bodies in reduced coordinates and x0-is mid point of x
// chZ --- mean Value evaluation of freduced(bodies),
{

  int dimRCS = bodies.dimRedConfSpace();

  MyVector Fx0(dimRCS);
  MyVector x0(dimRCS), x(dimRCS);
  MyVector xMinusx0(dimRCS);

  Bodies *midBodies = bodies.calculateMidBodies();
  x0 = midBodies->toReducedSpace();
  midBodies->compFRed(Fx0);
  delete midBodies;

  MyMatrix CDFF = -C * DFF;
  for (int i = 0; i < dimRCS; i++)
    CDFF[i][i] += 1.0;
  // now CDFF = I-C*DFF

  x = bodies.toReducedSpace();
  xMinusx0 = x - x0;

  chZ = Fx0 + DFF * xMinusx0; // Mean Value Form

  return x0 - C * Fx0 + CDFF * xMinusx0;
}

//*/
//-----------------------------------------------------------------------------------
ResultType krawczykMethod(Bodies boxBodies, Bodies &returnValues, bool &intersec, MyMatrix &DFF, bool recomputeDFF , bool thisIsUniqueZero )
  // the Krawczyk method is applied to <boxBodies>
  // On input: DFF = DfRed(boxbodies) - it is supplied because this is a costly function and is used elsewhere
  //     recomputeDFF - tells whether in each iteration we compute DFF, otherwise we just use DFF supplied on input
  //     if thisIsUniqueZero == false, then it is a normal run of the Krawczyk operator , we iterate ... and check the inclusion
  //        in the interior to have the proof of the existence of the solution
  //     if thisIsUniqueZero == true, then we KNOW from somewhere that <boxBodies> contains a unique zero
  //        and we just use the Krawczyk operator in order to refine the bounds
  ////////////
  //   one reason for the parameter "thisIsUniqueZero" is the following observation ; when iterating a box with a unique solution 
  //   for very many iterates it may happen that the box stabilizes and no longer we have that Krawczyk(box) \subset int box
  //     still in such case we do not want return the message that the method failed  
  // On output:
  //    intersec = true  when the method fails but still KrawczykMethod succeded in reducing the size of the search box
  //               false otherwise
  //   returnValues - is the result of Newton/Krawczyk operator interations (Nx \cap x if the intersection happens to be non-empty)
  //    DFF contains DfRed(returnValues) if recomputeDFF==true
{
  
  int max_iter=MAX_ITER_NEWTON;
  if(thisIsUniqueZero)
    max_iter=50;
  
  
  int dimRCS = boxBodies.dimRedConfSpace();
  MyMatrix C(dimRCS, dimRCS);
  MyVector Nx(dimRCS), Y(dimRCS), chZ(dimRCS);
  MyDouble accuracy = 1e-15; // proceed until the interval is sufficiently small

  MyDouble volboxBodies;

  bool foundSolution = false;
  intersec = false;
  if (thisIsUniqueZero){
    foundSolution = true;
    intersec = true;
  }


#ifdef TESTY_KRAWCZYK
  cout << "=======================================================================" << endl;
  cout << "po wejsciu do metody Krawczyka " << endl;
  cout << "dimRCS=" << dimRCS << endl;
  MyVector xboxb(dimRCS), Fxboxb(dimRCS);
  xboxb = boxBodies.toReducedSpace();
  boxBodies.compFRed(Fxboxb);
  cout << "xboxb=" << xboxb << endl;
  cout << "FRed(xboxb)=" << Fxboxb << endl;
  cout << endl;
#endif

  //  Computation of matrix C
  DMatrix midDF(dimRCS, dimRCS);
  midDF = capd::vectalg::convertObject<DMatrix>(DFF);
  
  bool C_comp_ok=true;   // used to tell whether the computation of C matrix succeded
  try
  {
    DMatrix doubleC(dimRCS, dimRCS);
    doubleC = capd::matrixAlgorithms::gaussInverseMatrix(midDF);
    C = capd::vectalg::convertObject<MyMatrix>(doubleC);
    if (!isomorphismTest(C))
	   C_comp_ok=false;
  }
  catch (std::runtime_error)
  {
     C_comp_ok=false;
  }
  if(! C_comp_ok){
    #ifdef REFINE_CC_TESTY
    cout << "Failure to compute C matrix " << endl;
	#endif
	if(thisIsUniqueZero)
	  return ResultType::uniqueZero;
	else
      return ResultType::methodFailed;
  }	
	
	
  // now C\appox mid(DFF)^{-1}

  bool finished = false;
  int loopC = 0;
  while (!finished)
  {
    Nx = krawczykOperator(C, DFF, boxBodies, chZ);

#ifdef TESTY_KRAWCZYK
    cout << "Result of KrawczykOperator:" << endl;
    cout << Nx << endl;
    cout << endl;
#endif

    for (int i = 0; i < dimRCS - 1; i++) // we check whether the mean value evaluation performed in <krawczykOperator> shows that there is no zero in <boxBodies>
    {
      if (!isSingular(chZ[i]))
      {
        finished = true;
        // retVal=ResultType::noZeroInSet;
        return ResultType::noZeroInSet; // there is no solution
      }
    }

    if (!finished)
    {
      if (cubeSubSetInterior(Nx, boxBodies))
      {
        foundSolution = true;
      }
      else if (cubeEmptyIntersection(Nx, boxBodies))
      {
        finished = true;
        // retVal=ResultType::noZeroInSet;
        return ResultType::noZeroInSet; // there is no solution
      }
    }
    if (!finished)
    {
      if (cubeSuperSet(Nx, boxBodies)) // bodies \subset Nx
      {
        finished = true;
        returnValues = boxBodies;
        // retval=foundSolution ? ResultType::uniqueZero : ResultType::methodFailed;
        return foundSolution ? ResultType::uniqueZero : ResultType::methodFailed;
      }
    }
    if (!finished)
    {
      intersec = true;

      volboxBodies = volume(boxBodies);

      Y = cubeIntersection(Nx, boxBodies);
      updateBodies(Y, boxBodies); // boxBodies=Y (in the reduced configuration space)

      // now we decide if it is worth to continue with the iteration
      MyDouble v = volume(boxBodies);
      if ((!thisIsUniqueZero) && ((1.0 - v / volboxBodies < 0.8) || (power(v, 1.0 / dimRCS) <= accuracy)) )
      {
        finished = true;
        returnValues = boxBodies;
        return foundSolution ? ResultType::uniqueZero : ResultType::methodFailed;
      }

      
      if (recomputeDFF)
        computeDFRedMatrix(boxBodies, DFF);

#ifdef TESTY_KRAWCZYK
      //boxBodies.printDiam();
      int maxi, coord;
      cout << "max diam of boxBodies " << boxBodies.maxDiam(maxi, coord) << endl;
      int max_i, max_j;
      cout << "max diam of DFF " << maxDiamOfMatrix(DFF, dimRCS, max_i, max_j) << endl;
      cout << "max_i = " << max_i << ", max_j = " << max_j << endl;
#endif

    } // if !finished
    loopC++;
    finished = finished || (loopC >= max_iter);
  } // end while (!finished)

  returnValues = boxBodies;
  return foundSolution ? ResultType::uniqueZero : ResultType::methodFailed;
}



ResultType refineCC(Bodies &bodies)
// the Krawczyk method is applied to a box containing CC in order to refine bounds
{
    ResultType newtonRes = ResultType::uniqueZero;
	#ifdef REFINE_CC_TESTY
    cout << "In refineCC... before krawczykMethod " << endl;
    bodies.printData();
    cout << endl;
    bodies.printDiam();
    cout << endl;
    #endif
  
  
    bool intersec = false; // when true then we know that we got an improvement in Newton/Krawczyk method
    int dimRCS = bodies.dimRedConfSpace();
    MyMatrix DFF(dimRCS, dimRCS);
    Bodies bodiesRetFromKraw(bodies.N);
    computeDFRedMatrix(bodies, DFF);

    bool recomputeDFF=true;
	bool thisIsBoxWithUniqueSol=true;
    newtonRes = krawczykMethod(bodies, bodiesRetFromKraw, intersec, DFF, recomputeDFF,thisIsBoxWithUniqueSol);
    bodies = bodiesRetFromKraw;
	#ifdef REFINE_CC_TESTY
    cout << "in refineCC... after krawczykMethod " << endl;
    bodies.printData();
    cout << endl;
    bodies.printDiam();
    cout << endl;
	#endif

    if (newtonRes != ResultType::uniqueZero)
    {

        if (newtonRes == ResultType::methodFailed)
        {
            cout << "methodFailed" << endl;
        }
        if (newtonRes == ResultType::noZeroInSet)
        {
            cout << "noZeroInSet" << endl;
        }
        cout << "DISASTER!" << endl;
		
        exit(4);
    }

    return newtonRes;
}